#include "xprt.h"

/* --------------------------- GLOBAL VARIABLES ----------------------- */

#define __GLOBALS__

InputVector *gX; 		
ExampleVector *gY;
CoordinateNumber gDimX, gDimY;  
ExampleNumber gNumPts;          
ModelType gXprtModel;

/* ------------------------------- main ------------------------------- */

main(int argc, char **argv)
{
   Node *tree;
   int epoch;

   tree = SetUp(argc, argv);  /* initialize globals and tree */
   printf("Initial error:\t");
   NetTest(tree);

   for (epoch=1; epoch<=kNumEpochs; ++epoch)
   {
      free(Compute(tree)); /* compute probabilities and expert guesses  */
      Update(tree, NULL);  /* update weight vectors and variances       */ 

      printf("Epoch %d:\t", epoch);
      NetTest(tree);
   }
} 

/* ------------------------------- Setup ------------------------------ */

Node *SetUp(int argc, char **argv)
{
   char fname[kStrLen], line[kStrLen];
   FILE *fpParam, *fpData;
   Node *tree;
   CoordinateNumber n;
   ExampleNumber t;
   int nSplits;

   if (argc != kNumArgs)
      Error("usage -- xprt <infile>");
   fpParam = SafeOpen(argv[1], "r");

   /***** create input/output data matrix  *****/

   fgets(line, kStrLen, fpParam); 
   sscanf(line, "%d%d%d%d", &gDimX, &gDimY, &gNumPts, &gXprtModel);

   gX = NewMatrix(gNumPts, ++gDimX); /* bias adds extra coordinate */
   gY = NewMatrix(gDimY, gNumPts);

   /***** read training data *****/

   fgets(line, kStrLen, fpParam); 
   sscanf(line, "%s", fname);   

   fpData = SafeOpen(fname, "r");
   for (t=0; t<gNumPts; ++t)
   {
      gX[t][0] = 1.0; /* bias coordinate */
      for (n=1; n<gDimX; ++n)
         fscanf(fpData, "%lf", &gX[t][n]); 
      for (n=0; n<gDimY; ++n)
         fscanf(fpData, "%lf", &gY[n][t]); 
   }
   fclose(fpData);

   /***** create tree *****/

   tree = NewNode();
   tree->xprt = NewXprt();

   fgets(line, kStrLen, fpParam); 
   sscanf(line, "%d", &nSplits);

   while (nSplits)
   {
      AddLayer(tree, &nSplits);
      fgets(line, kStrLen, fpParam); 
      sscanf(line, "%d", &nSplits);
   }

   fclose(fpParam);
   return (tree);
}

/* ------------------------------- Compute ------------------------------- */

ExampleVector Compute(Node *node)
{
   ExampleVector p, p1, p2; 
   ExampleNumber t;
   CoordinateNumber i, j;
   double u; 
   Xprt *xprt;
   Gate *gate;

   p = NewVector(gNumPts);

   if (xprt=IsXprt(node))
   {
      switch (gXprtModel) {
         case gaussian:
            for (t=0; t<gNumPts; ++t)
            {
               xprt->var[t] = 0.0;
               for (i=0; i<gDimY; ++i)
               {
                  for (j=0, u=gY[i][t]; j<gDimX; ++j)
                     u -= xprt->wt[i][j] * gX[t][j];
                  xprt->var[t] += u*u;
               }
               p[t] = exp(-xprt->var[t]/(2.0*xprt->phi)) /
                      pow(xprt->phi, gDimY/2.0);
            }
            break;
         case bernoulli:
            for (t=0; t<gNumPts; ++t)
            {
               p[t] = 1.0;
               for (i=0; i<gDimY; ++i)
               {
                  for (j=0, u=0.0; j<gDimX; ++j)
                     u += gX[t][j] * xprt->wt[i][j];
                  u = 1.0/(1.0 + exp(-u));
                  p[t] *= (gY[i][t] ? u : 1.0-u);
               }
            }
            break;
      }
   }
   else
   if (gate=IsGate(node))
   {
      p1 = Compute(gate->node1);
      p2 = Compute(gate->node2);
      
      for (t=0; t<gNumPts; ++t)
      {
         for (i=0, u=0.0; i<gDimX; ++i)
            u += gX[t][i] * gate->wt[i];
         u = 1.0/(1.0 + exp(-u));
         p[t] = u*p1[t] + (1.0-u)*p2[t]; 
         gate->h[t] = u*p1[t]/p[t]; 
      }

      free(p1); free(p2);
   }

   return (p);
}

/* ------------------------------- Update ------------------------------ */

void Update(Node *node, ExampleVector hp)
{
   ExampleVector hc;
   CoordinateNumber n;
   ExampleNumber t;
   double norm;
   Xprt *xprt;
   Gate *gate;
   
   if (xprt=IsXprt(node))
   {
      if (gXprtModel==gaussian)
      {
         if (hp==NULL) /* the tree has no gating networks */ 
         {
            for (t=0; t<gNumPts; ++t)
               xprt->phi += xprt->var[t];
            xprt->phi /= gNumPts;
         }
         else
         { 
            norm = xprt->phi = 0.0;
            for (t=0; t<gNumPts; ++t)
            {
               xprt->phi += hp[t] * xprt->var[t];
               norm += hp[t];
            }             
            if (norm)
               xprt->phi /= norm;
            else
               xprt->phi = 1.0;
         }
      }

      for (n=0; n<gDimY; ++n)
         IRLS(gY[n], hp, xprt->wt[n], gXprtModel, xprt->phi);
   }
   else
   if (gate=IsGate(node))
   {
      IRLS(gate->h, hp, gate->wt, bernoulli, kBernoulliPhi);
      hc = NewVector(gNumPts);

      if (hp==NULL)
      {
         Update(gate->node1, gate->h);

         for (t=0; t<gNumPts; ++t)
            hc[t] = 1.0 - gate->h[t];
         Update(gate->node2, hc);
      }
      else
      {
         for (t=0; t<gNumPts; ++t)
            hc[t] = hp[t] * gate->h[t];
         Update(gate->node1, hc);

         for (t=0; t<gNumPts; ++t)
            hc[t] = hp[t] * (1.0-gate->h[t]);
         Update(gate->node2, hc);
      }

      free(hc);
   }
}

/* ------------------------------ NetOut --------------------------------- */

void NetOut(Node *node, InputVector x, OutputVector y)
{ 
   OutputVector y1, y2;
   CoordinateNumber m, n;
   double p;
   Xprt *xprt;
   Gate *gate;

   if (xprt=IsXprt(node))
   {
      switch (gXprtModel) {
         case gaussian:
            for (m=0; m<gDimY; ++m)
               for (n=0, y[m] = 0.0; n<gDimX; ++n)
                  y[m] += xprt->wt[m][n] * x[n]; 
            break;
         case bernoulli:
            for (m=0; m<gDimY; ++m)
            {
               p = 0.0;
               for (n=0; n<gDimX; ++n)
                  p += xprt->wt[m][n] * x[n];
               y[m] = 1.0/(1.0 + exp(-p));
            }
            break;
      }
   }
   else
   if (gate=IsGate(node))
   {
      y1 = NewVector(gDimY);
      y2 = NewVector(gDimY);
       
      NetOut(gate->node1, x, y1);
      NetOut(gate->node2, x, y2);
 
      for (n=0, p=0.0; n<gDimX; ++n)
         p += x[n] * gate->wt[n];

      p = 1.0/(1.0+exp(-p));

      for (n=0; n<gDimY; ++n)
         y[n] = p*y1[n] + (1.0-p)*y2[n];

      free(y1);
      free(y2);
   }
}

/* ------------------------------ NetTest -------------------------------- */

void NetTest(Node *tree)
{
   Vector y;
   ExampleNumber t;
   CoordinateNumber n;
   double dy, err=0.0;

   y = NewVector(gDimY);

   for (t=0; t<gNumPts; ++t)
   {
      NetOut(tree, gX[t], y);
      for (n=0; n<gDimY; ++n)
      {
         dy = (y[n] - gY[n][t]);
         err += dy * dy; 
      }
   }

   printf("err = %f\n", err);
}

/* -------------------------------------------------------------------- */
